@@ -507,7 +507,20 @@ class BrokerConnection(object):
             return None
 
         # decode response
-        response = ifr.response_type.decode(read_buffer)
+        try:
+            response = ifr.response_type.decode(read_buffer)
+        except ValueError:
+            read_buffer.seek(0)
+            buf = read_buffer.read()
+            log.error('%s Response %d [ResponseType: %s Request: %s]:'
+                      ' Unable to decode %d-byte buffer: %r', self,
+                      ifr.correlation_id, ifr.response_type,
+                      ifr.request, len(buf), buf)
+            ifr.future.failure(Errors.UnknownError('Unable to decode response'))
+            self.close()
+            self._processing = False
+            return None
+
         log.debug('%s Response %d: %s', self, ifr.correlation_id, response)
         ifr.future.success(response)
         self._processing = False
@@ -548,9 +561,9 @@ class BrokerConnection(object):
 
         class ConnFilter(Filter):
             def filter(self, record):
-                if record.funcName in ('recv', 'send'):
-                    return False
-                return True
+                if record.funcName == 'check_version':
+                    return True
+                return False
         log_filter = ConnFilter()
         log.addFilter(log_filter)
 
@@ -585,11 +598,11 @@ class BrokerConnection(object):
             # the attempt to write to a disconnected socket should
             # immediately fail and allow us to infer that the prior
             # request was unrecognized
-            self.send(MetadataRequest[0]([]))
+            mr = self.send(MetadataRequest[0]([]))
 
             if self._sock:
                 self._sock.setblocking(True)
-            while not f.is_done:
+            while not (f.is_done and mr.is_done):
                 self.recv()
             if self._sock:
                 self._sock.setblocking(False)
